;*
;* Map Pan Routines
;*
;*   Started 15-Apr-97
;*
;* Initials: JF = Jeff Frohwein
;*
;* V1.0 -  5-May-97 : Original Release - JF
;* V1.1 -  6-May-97 : Optimized to make smaller - JF
;* V1.2 -  8-May-97 : Pan relatives set Carry if fail - JF
;* V1.3 - 12-Jun-97 : Modified for RGBASM - JF
;* V1.4 - 18-Jul-97 : Made includeable or compileable - JF
;* V1.5 - 21-Jul-97 : Supports latest 'shift.inc' - JF
;* V1.6 - 27-Jul-97 : Supports new subroutine prefixes - JF
;* V1.7 - 17-Aug-97 : No longer do DEF on Multiply. (Bad!) - JF
;*
;* Library Subroutines:
;*
;* pan_Init -
;*   Initialize Map Location
;*    Entry: hl = map address
;*           a = screen address (0 = $9800, other = $9c00)
;*
;* pan_Absolute -
;*   Draw map at X,Y
;*    Entry: de = X, hl = Y
;*    (If hl or de too large, stop at boundaries.)
;*
;* pan_Left -
;*   Pan left one pixel.
;*   (Don't pan if at boundary & set Carry flag.)
;*
;* pan_Right -
;*   Pan right one pixel.
;*   (Don't pan if at boundary & set Carry flag.)
;*
;* pan_Up -
;*   Pan up one pixel.
;*   (Don't pan if at boundary & set Carry flag.)
;*
;* pan_Down -
;*   Pan down one pixel.
;*   (Don't pan if at boundary & set Carry flag.)
;*
;* pan_UpRight -
;*   Pan up & right one pixel.
;*
;* pan_UpLeft -
;*   Pan up & left one pixel.
;*
;* pan_DownRight -
;*   Pan down & right one pixel.
;*
;* pan_DownLeft -
;*   Pan down & left one pixel.
;*
;* pan_Coord -
;*   Return screen upper left coords relative to map.
;*    Exit: de = X, hl = Y
;*
;* definitions:
;*
;*  map address -
;*
;*   This is a pointer to map data with following format:
;*    word map_Width (in bytes not pixels)
;*    word map_Height (in bytes not pixels)
;*    map data - Starting from upper left corner of map.
;*     row1/row2/row3/etc. (contains map_Width * map_Height bytes)

; If all of these are already defined, don't do it again.

        IF      !DEF(PAN1_ASM)
PAN1_ASM  SET  1

rev_Check_pan1_asm: MACRO
;NOTE: REVISION NUMBER CHANGES MUST BE ADDED
;TO SECOND PARAMETER IN FOLLOWING LINE.
        IF      \1 > 1.7      ; <--- PUT REVISION NUMBER HERE
        WARN    "Version \1 or later of 'pan1.asm' is required."
        ENDC
        ENDM

; Definition includes

        INCLUDE "hardware.inc"          ;Hardware defines
        INCLUDE "incdec.inc"            ;Sprite trash prevention macros
        INCLUDE "shift.inc"             ;Shift math

Multiply EQUS "mul_DE_HL_BCDEu"

; Code includes

        INCLUDE "memory1.asm"

        IF      DEF(U161632A_ASM)

         GLOBAL  Multiply               ;Multiply is already defined.

        ELSE

         INCLUDE "u161632a.asm"         ;Include Multiply code.

        ENDC

; Make sure include files are useable revisions.

        rev_Check_hardware_inc   1.5
        rev_Check_incdec_inc     1.1
        rev_Check_memory1_asm    1.0
        rev_Check_shift_inc      1.2

;Low RAM Assignments

        SECTION "Pan1 Ram", BSS

MapAddr         dw           ; address of map in memory
MapULX          dw           ; X coord of upper left of pan window
MapULY          dw           ; Y coord of upper left of pan window
MapDimX         dw           ; width of map in pixels
MapDimY         dw           ; height of map in pixels
MapFirstTime    db           ; First time map drawn (0 = no, 1 = yes)
ScrnAddr        db           ; screen memory address high byte

        SECTION "Pan1 Code", HOME

;* Initialize Map Location *
; Entry: hl = map address
;        a = screen address (0 = $9800, other = $9c00)

pan_Init::
        or      a               ;is it $9800?
        ld      a,$98
        jr      z,pani1         ;yes
        ld      a,$9c
pani1:  ld      [ScrnAddr],a    ;save screen address high byte

        ld      a,[hl+]         ;save map width
        ld      [MapDimX],a
        ld      a,[hl+]
        ld      [MapDimX+1],a

        ld      a,[hl+]         ;save map height
        ld      [MapDimY],a
        ld      a,[hl+]
        ld      [MapDimY+1],a

        ld      a,l             ;save map address
        ld      [MapAddr],a
        ld      a,h
        ld      [MapAddr+1],a

        xor     a
        ld      [MapULX],a
        ld      [MapULX+1],a
        ld      [MapULY],a
        ld      [MapULY+1],a

        ldh     [rSCY],a        ;screen scroll Y=0
        ldh     [rSCX],a        ;screen scroll X=0

        inc     a
        ld      [MapFirstTime],a        ;flag to draw everything
        ret

;* Return pan coordinates in DE,HL

pan_Coord::
        ld      a,[MapULX]
        ld      e,a
        ld      a,[MapULX+1]
        ld      d,a
        ld      a,[MapULY]
        ld      l,a
        ld      a,[MapULY+1]
        ld      h,a
        ret

; Draw map at X,Y (DE,HL)

pan_Absolute::
        ld      a,[MapFirstTime]
        or      a                       ;Has map completely been drawn once?

        push    hl
        push    de

        call    nz,DrawOnce             ;no
        pop     hl

        ld      a,[MapULX]
        ld      c,a
        ld      a,[MapULX+1]
        ld      b,a
        call    mapNegate

        add     hl,bc
        ld      b,h
        ld      c,l

        ld      a,b
        or      c                       ;Any change in X coord?
        jr      z,pana3                 ;no

        bit     7,h                     ;Should we scroll left or right?
        jr      z,pana2                 ;right

        call    mapNegate
pana1:  push    bc
        call    pan_Left
        pop     bc
        dec     bc
        ld      a,b
        or      c
        jr      nz,pana1

        jr      pana3

pana2:  push    bc
        call    pan_Right
        pop     bc
        dec     bc
        ld      a,b
        or      c
        jr      nz,pana2

pana3:
        pop     hl

        ld      a,[MapULY]
        ld      c,a
        ld      a,[MapULY+1]
        ld      b,a
        call    mapNegate

        add     hl,bc
        ld      b,h
        ld      c,l

        ld      a,b
        or      c                       ;Any change in Y coord?
        jr      z,pana6                 ;no

        bit     7,h                     ;Should we scroll up or down?
        jr      z,pana5                 ;down

        call    mapNegate
pana4:  push    bc
        call    pan_Up
        pop     bc
        dec     bc
        ld      a,b
        or      c
        jr      nz,pana4

        jr      pana6

pana5:  push    bc
        call    pan_Down
        pop     bc
        dec     bc
        ld      a,b
        or      c
        jr      nz,pana5
pana6:
        ret

mapNegate:
        ld      a,c
        cpl
        ld      c,a
        ld      a,b
        cpl
        ld      b,a
        INC_BC
        ret

; Draw map at 0,0 one time
DrawOnce:
        ld      a,[MapAddr]
        ld      c,a
        ld      a,[MapAddr+1]
        ld      b,a

        ld      a,[ScrnAddr]
        ld      h,a
        ld      l,0

        ld     e,32
pana8:
        ld     d,32
        push   bc
pana9:
        di
        lcd_WaitVRAM
        ld     a,[bc]
        ld     [hl+],a
        ei
        inc     bc
        dec     d
        jr      nz,pana9

        pop     bc

; bc = bc + MapDimX
        push    hl
        ld      a,[MapDimX]
        ld      l,a
        ld      a,[MapDimX+1]
        ld      h,a
        add     hl,bc
        ld      c,l
        ld      b,h
        pop     hl

        dec    e
        jr     nz,pana8

        xor     a
        ld      [MapFirstTime],a        ;Indicate map drawn once
        ret

;* Pan right test *
;Exit: Set carry if pan right not possible

pan_RightTest:
        ld      a,[MapULX+1]    ; Increment absolute X coord
        ld      h,a
        ld      a,[MapULX]
        ld      l,a

        push    hl              ; scroll attempted too far right?

        ld      a,[MapDimX]
        ld      e,a
        ld      a,[MapDimX+1]
        ld      d,a
        ld      hl,-SCRN_X_B
        add     hl,de
        add     hl,hl
        add     hl,hl
        add     hl,hl

        dec     hl
        ld      a,h
        cpl
        ld      d,a
        ld      a,l
        cpl
        ld      e,a             ; de = -[[MapDimX - SCRN_X_B] * 8]

        pop     hl

        add     hl,de
        ld      a,h
        and     $80
        scf
        ret     z               ; hit boundary, return
        or      a               ;clear carry
        ret

;* Pan left test *
;Exit: Set carry if pan left not possible

pan_LeftTest:
        ld      a,[MapULX+1]    ; Decrement absolute X coord
        ld      h,a
        ld      a,[MapULX]
        ld      l,a
        or      h               ; Are we at left boundary?
        scf
        ret     z               ; yes, don't allow scroll to occur
        or      a
        ret

;* Pan down test *
;Exit: Set carry if pan down not possible

pan_DownTest:
        ld      a,[MapULY+1]    ; Increment absolute Y coord
        ld      h,a
        ld      a,[MapULY]
        ld      l,a

        push    hl              ; scroll attempted too far down?

        ld      a,[MapDimY]
        ld      e,a
        ld      a,[MapDimY+1]
        ld      d,a
        ld      hl,-SCRN_Y_B
        add     hl,de
        add     hl,hl
        add     hl,hl
        add     hl,hl

        dec     hl
        ld      a,h
        cpl
        ld      d,a
        ld      a,l
        cpl
        ld      e,a      ; de = -[[MapDimY - SCRN_Y_B] * 8]

        pop     hl

        add     hl,de
        ld      a,h
        and     $80
        scf
        ret     z               ; hit boundary, return
        or      a
        ret

;* Pan up test *
;Exit: Set carry if pan up not possible

pan_UpTest:
        ld      a,[MapULY+1]    ; Increment absolute Y coord
        ld      h,a
        ld      a,[MapULY]
        ld      l,a
        or      h               ; Are we at top boundary?
        scf
        ret     z               ; yes, don't allow scroll to occur
        or      a
        ret

;* Pan up & right on map 1 dot *

pan_UpRight::
        call    pan_UpTest      ; Is pan up possible?
        ret     c               ; no
        call    pan_RightTest   ; Is pan right possible?
        ret     c               ; no, hit boundary, return
        call    UpRaw
        call    RightRaw
        or      a
        ret

;* Pan up & left on map 1 dot *

pan_UpLeft::
        call    pan_UpTest      ; Is pan up possible?
        ret     c               ; no
        call    pan_LeftTest    ; Is pan left possible?
        ret     c               ; no, hit boundary, return
        call    UpRaw
        call    LeftRaw
        or      a
        ret

;* Pan down & right on map 1 dot *

pan_DownRight::
        call    pan_DownTest    ; Is pan down possible?
        ret     c               ; no
        call    pan_RightTest   ; Is pan right possible?
        ret     c               ; no, hit boundary, return
        call    DownRaw
        call    RightRaw
        or      a
        ret

;* Pan down & left on map 1 dot *

pan_DownLeft::
        call    pan_DownTest    ; Is pan down possible?
        ret     c               ; no
        call    pan_LeftTest    ; Is pan left possible?
        ret     c               ; no, hit boundary, return
        call    DownRaw
        call    LeftRaw
        or      a
        ret

;* Pan right on map 1 dot *

pan_Right::
        call    pan_RightTest   ; Is pan right possible?
        ret     c               ; no, hit boundary, return

RightRaw:
        ld      a,[MapULX+1]    ; Increment absolute X coord
        ld      h,a
        ld      a,[MapULX]
        ld      l,a

        inc     hl

        ld      a,h
        ld      [MapULX+1],a
        ld      a,l
        ld      [MapULX],a
        dec     a
        and     7        ;Do we need to load bytes on unseen right column?
        jp      nz,PanLRX ;no

;Fill far right currently unseen column with map data

        srl16   hl,3       ;divide hl by 8

        ld      bc,SCRN_X_B
        add     hl,bc
        push    hl

        ld      a,[MapULY]
        ld      l,a
        ld      a,[MapULY+1]
        ld      h,a

        srl16   hl,3       ;divide hl by 8

        ld      a,[MapDimX]
        ld      e,a
        ld      a,[MapDimX+1]
        ld      d,a

        call    Multiply        ;de = [MapULY/8] * MapDimX

        ld      a,[MapDimX]
        cpl
        ld      l,a
        ld      a,[MapDimX+1]
        cpl
        ld      h,a
        ld      bc,1            ; CAN'T USE 'INC HL' HERE
        add     hl,bc           ;hl = - MapDimX

        add     hl,de
        ld      d,h
        ld      e,l             ;de = [[MapULY/8] * MapDimX] - MapDimX

        pop     hl              ;hl = [MapULX/8] + SCRN_X_B
        push    hl

        add     hl,de
        ld      a,[MapAddr]
        ld      e,a
        ld      a,[MapAddr+1]
        ld      d,a
        add     hl,de
        ld      b,h             ;bc = MapAddr + [MapULX/8] +
        ld      c,l             ;     [[MapULY/8] * NumCols] -
                                ;     NumCols + SCRN_X_B
pan_lr:
        ld      a,[MapULY]
        sub     8
        and     $f8
        ld      l,a
        ld      h,0
        add     hl,hl
        add     hl,hl           ;hl = [[[MapULY-8] and 00f8]*4]

        pop     de              ;de = [MapULX/8] + SCRN_X_B
        ld      a,e
        and     31
        ld      e,a
        ld      d,0
        add     hl,de
        ld      e,d
        ld      a,[ScrnAddr]
        ld      d,a
        add     hl,de           ;hl = ScrnAddr + [[[MapULY-8] and 00f8]*4] +
                                ;     [[[MapULX/8] + SCRN_X_B] and 31]
        ld      d,SCRN_Y_B+2
PanR1:
        push    de
        di
        lcd_WaitVRAM
        ld     a,[bc]
        ld     [hl],a
        ei

; increment down in map data

        push    hl              ;bc = bc + MapDimX
        ld      a,[MapDimX]
        ld      l,a
        ld      a,[MapDimX+1]
        ld      h,a
        add     hl,bc
        ld      b,h
        ld      c,l
        pop     hl

; increment down in screen memory, wrap around if necessary

        ld      de,32           ;hl = [[hl + 32] and $bff] or ScrnAddrH
        add     hl,de
        ld      a,[ScrnAddr]
        ld      d,a
        ld      a,h
        and     $0b
        or      d
        ld      h,a

        pop     de
        dec     d
        jr     nz,PanR1
PanLRX:
        ld      a,[MapULX]      ;scroll screen right 1 dot
        ldh     [rSCX],a
        or      a               ;clear carry flag
        ret

;* Pan left on map 1 dot *

pan_Left::
        call    pan_LeftTest    ; Is pan left possible?
        ret     c               ; no

LeftRaw:
        ld      a,[MapULX+1]    ; Decrement absolute X coord
        ld      h,a
        ld      a,[MapULX]
        ld      l,a

        dec     hl

        ld      a,h
        ld      [MapULX+1],a
        ld      a,l
        ld      [MapULX],a
        inc     a
        and     7        ;Do we need to load bytes on unseen left column?
        jp      nz,PanLRX ;no

;Fill far left currently unseen column with map data

        srl16   hl,3       ;divide hl by 8

        dec     hl
        push    hl

        ld      a,[MapULY]
        ld      l,a
        ld      a,[MapULY+1]
        ld      h,a

        srl16   hl,3       ;divide hl by 8

        ld      a,[MapDimX]
        ld      e,a
        ld      a,[MapDimX+1]
        ld      d,a

        call    Multiply        ;de = [MapULY/8] * MapDimX

        ld      a,[MapDimX]
        cpl
        ld      l,a
        ld      a,[MapDimX+1]
        cpl
        ld      h,a
        ld      bc,1            ;CAN'T USE 'INC HL' HERE
        add     hl,bc           ;hl = - MapDimX

        add     hl,de
        ld      d,h
        ld      e,l             ;de = [[MapULY/8] * MapDimX] - MapDimX

        pop     hl              ;hl = [MapULX/8] - 1
        push    hl

        add     hl,de
        ld      a,[MapAddr]
        ld      e,a
        ld      a,[MapAddr+1]
        ld      d,a
        add     hl,de
        ld      b,h             ;bc = MapAddr + [MapULX/8] +
        ld      c,l             ;     [[MapULY/8] * NumCols] - 1

        jp      pan_lr

;* Pan down on map 1 dot *

pan_Down::
        call    pan_DownTest    ; is panning down possible?
        ret     c               ; no, border hit

DownRaw:
        ld      a,[MapULY+1]    ; Increment absolute Y coord
        ld      h,a
        ld      a,[MapULY]
        ld      l,a

        inc     hl

        ld      a,h
        ld      [MapULY+1],a
        ld      a,l
        ld      [MapULY],a
        dec     a
        and     7        ;Do we need too load bytes on unseen bottom row?
        jp      nz,PanUDX ;no

;Fill far bottom currently unseen row with map data

        srl16   hl,3       ;divide hl by 8

        ld      bc,SCRN_Y_B
        add     hl,bc
        push    hl

        ld      a,[MapDimX]
        ld      e,a
        ld      a,[MapDimX+1]
        ld      d,a

        call    Multiply        ;de = [[MapULY/8]+18] * NumCols

        ld      a,[MapULX]
        ld      l,a
        ld      a,[MapULX+1]
        ld      h,a

        srl16   hl,3            ;divide hl by 8

        add     hl,de           ;hl = [MapULX/8] + [[[MapULY/8]+18] * NumCols]

        ld      a,[MapAddr]
        ld      e,a
        ld      a,[MapAddr+1]
        ld      d,a
        add     hl,de
        ld      b,h             ;bc = MapAddr + [MapULX/8] +
        ld      c,l             ;     [[[MapULY/8]+18] * NumCols] - 1
        dec     bc

        pop     hl              ;hl = [MapULY/8] + 18
        add     hl,hl
        add     hl,hl
        add     hl,hl
        ld      h,0
        add     hl,hl
        add     hl,hl           ;hl = [[[[MapULY/8] + 18] * 32] and 3ff]
pan_ud:
        ld      a,[MapULX]
        ld      e,a
        ld      a,[MapULX+1]
        ld      d,a

        srl16   de,3            ;divide de by 8
        dec     de

        ld      a,e
        and     $1f
        ld      e,a
        ld      d,0             ;de = [[MapULX/8]-1] and 31

        add     hl,de
        ld      e,d
        ld      a,[ScrnAddr]
        ld      d,a
        add     hl,de           ;hl = ScrnAddr + [[[MapULX/8]-1] and 31] +
                                ;     [[[[MapULY/8]+18]*32] and 3ff]

        ld      d,SCRN_X_B+2
PanD1:
        di
        lcd_WaitVRAM
        ld     a,[bc]
        ld     [hl+],a
        ei

; increment right in map data

        inc     bc

; wrap around in screen memory if necessary

        ld      a,l
        and     $1f
        jr      nz,PanD2
        dec     hl
        ld      a,l
        and     $e0
        ld      l,a
PanD2:

        dec     d
        jr     nz,PanD1
PanUDX:
        ld      a,[MapULY]      ;scroll screen down 1 dot
        ldh     [rSCY],a
        or      a               ;clear carry flag
        ret

;* Pan up on map 1 dot *

pan_Up::
        call    pan_UpTest      ; Is panning up possible?
        ret     c               ; no

UpRaw:
        ld      a,[MapULY+1]    ; Increment absolute Y coord
        ld      h,a
        ld      a,[MapULY]
        ld      l,a

        dec     hl

        ld      a,h
        ld      [MapULY+1],a
        ld      a,l
        ld      [MapULY],a
        inc     a
        and     7        ;Do we need too load bytes on unseen top row?
        jp      nz,PanUDX ;no

;Fill far top currently unseen row with map data

        srl16   hl,3       ;divide hl by 8

        push    hl

        ld      a,[MapDimX]
        ld      e,a
        ld      a,[MapDimX+1]
        ld      d,a

        call    Multiply        ;de = [MapULY/8] * NumCols

        ld      a,[MapDimX]
        cpl
        ld      l,a
        ld      a,[MapDimX+1]
        cpl
        ld      h,a
        ld      bc,1            ; CAN'T USE 'INC HL' HERE
        add     hl,bc           ;hl = - MapDimX

        add     hl,de
        ld      d,h
        ld      e,l             ;de = [[MapULY/8] * NumCols] - NumCols

        ld      a,[MapULX]
        ld      l,a
        ld      a,[MapULX+1]
        ld      h,a

        srl16   hl,3            ;divide hl by 8
        add     hl,de           ;hl = [MapULX/8] +
                                ;     [[MapULY/8] * NumCols] - NumCols
        ld      a,[MapAddr]
        ld      e,a
        ld      a,[MapAddr+1]
        ld      d,a
        add     hl,de
        ld      b,h
        ld      c,l
        dec     bc              ;bc = MapAddr + [MapULX/8] +
                                ;     [[MapULY/8] * NumCols] - NumCols - 1

        pop     hl              ;hl = [MapULY/8]
        dec     hl
        add     hl,hl
        add     hl,hl
        add     hl,hl
        ld      h,0
        add     hl,hl
        add     hl,hl           ;hl = [[[[MapULY/8]-1] * 32] and 3ff]

        jp      pan_ud

        ENDC    ;pan1_asm
